/*
* SPIN2.C: Original by Nissim Chudnoff: Modified by Jon Griffiths
*
* Compares Speed Of (in order):                                             
* screen_blit_fs_buffer()                                                   
* screen_blit_buff_to()                                                     
* screen_blit_buff_to() (Dirty Rectangles)                                   
* doesn't test the NC functions as sprites can be drawn outside of the buffer.
*
* Notice that as the number of sprites increases, the dirty rectangle method
* gets less and less efficient. Possible optimisations that would reduce this:
  1. Keep a count of rectangles actually used and only blit that many.
  2. When adding dirty rectangles to the list, combine rectangles together if
     they overlap (i.e avoid re-drawing over an already drawn rectangle on
     screen.
  3. When the list reaches a certain size, just scrap the reactangle adding
     and use a big blit.
  4. Clip the dirty rectangles to the screen and use NC blitting functions.
  5. Use the size of the frame of the sprite when adding a dirty rectangle,
     not SPR_MAX_X/Y as here.
*
* Without optimisation 2, using the dirty rectangles results in flicker
* when rectangles overlap because the same screen spot is being updated more
* than once per frame - looks nasty.
*/

#include <stdio.h>
#include <stdlib.h>
#include <jlib.h>

#define RANDOM(u) ( rand() % ( ((u) + 1)  ) )

void usage(void);

int main(int argc,char *argv[])
{
 int num_iterations,num_sprites,cnt=0,i,*cur,animation[6] = {0,1,2,3,4,5};
 char times=1;                             /* Counter For blit Type */
 sprite_system *demo_sprites;
 buffer_rec *buff;
 unsigned int start_time,end_time; /* timing */
 int *dirty_recs;                          /* (x1,y1,x2,y2) of dirty rects */
 UBYTE *pal;

 if(argc != 3){
    usage();
 }

 num_sprites=atoi(argv[1]);
 num_iterations=atoi(argv[2]);

 if((num_sprites<=0)){
    usage();
 }
 if((num_iterations<=0)){
    usage();
 }

 /* allocate space for storing our rectangle array */
 dirty_recs=(int *)malloc((num_sprites*4)*sizeof(int));

 /* init stuff*/
 buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
 demo_sprites = sprite_init(num_sprites,6);
 pal = pal_load("spin.pal");
 
 if( sprite_load("spin.spr",demo_sprites) != SUCCESS ){
    puts("error: Couldn't load sprites");
    exit(-1);
 }
 
 if((buff== NULL)||(demo_sprites== NULL)||(pal== NULL)||(dirty_recs== NULL)){
    puts("Initialisation error.");
    exit(-1);
 }
 
 screen_set_video_mode();
 screen_block_set_pal(pal);
 srand(time(NULL));
 uclock_init();  /* set up our timing */

 /* Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     dirty_recs[i*4]=x;
     dirty_recs[i*4+1]=y;
     dirty_recs[i*4+2]=x+SPR_MAX_X;
     dirty_recs[i*4+3]=y+SPR_MAX_Y;

     sprite_set_xy(demo_sprites,i,x,y);
     sprite_set_an_frame(demo_sprites,i,0);
     sprite_set_move_info(demo_sprites,i,0,0,0); /* not moving */
     sprite_set_anim_info(demo_sprites,i,1,6,animation);
     sprite_turn_on(demo_sprites,i);
 }

 fgetc(stdin);
 start_time=uclock_read();

 /* this is the main sprite drawing loop */
 for(times = 1;times <= 3;){
    buff_save_all_sprites(demo_sprites,buff);
    buff_draw_all_sprites(demo_sprites,buff);

    cur=dirty_recs; /* get a pointer to the start of our dirty rect array */

    switch(times){
     case 1:
        screen_blit_fs_buffer(buff); 
        break;
     case 2:
        screen_blit_buff_to(0,0,buff,0,0,SCREEN_MAX_X,SCREEN_MAX_Y);
        break;
     case 3:
        /* Draw each dirty rectangle to the screen */
        for(i=0;i<num_sprites;i++){
            screen_blit_buff_to(cur[0],cur[1],buff,cur[0],cur[1],cur[2],cur[3]);
            cur+=4;
        }
        break;
    }

    if(cnt == num_iterations){
       /* move to next method */
       cnt = 0; /* reset count */
       end_time=uclock_read();
       printf("Method %d: %d Iterations in %d usec.\n",times,num_iterations,end_time-start_time);
       times++;
       if(times==3){ /* time to try dirty rectangles? */
          buff_clear(buff); /* clear because we'll only update changed parts */
       }
       fgetc(stdin);
       start_time=uclock_read();
    }
    else{
      cnt++; /* do next iteration with this method */
    }

    buff_rest_all_sprites(demo_sprites,buff);
    sprite_update_all_anim_and_move(demo_sprites);
 }

 screen_restore_video_mode();
 return 0; 
}

void usage(void)
{
 puts("spin2.exe : Speed up blitting with dirty rectangles blits.");
 puts("usage: spin2.exe <num_sprites> <num_iterations>");
 puts("Output is on stdout, so pipe it to a file to view.");
 exit(0);
}

