/*
* SPIN3.C: Original by Nissim Chudnoff: Modified by Jon Griffiths
*
* This is an attempt at a dynamic(ish) dirty rectangle framework.
* Uses a list to hold the rectangles, but doesn't update the list.
*/

#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include <jlib.h>

#define RANDOM(u) ( rand() % ( ((u) + 1)  ) )

void usage(void);
typedef struct {
        int x1,y1,x2,y2;
} drect;

void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2);
int drect_do_intersect(drect *one,drect *two);

int main(int argc,char *argv[])
{
 int num_iterations,num_sprites,i,seed,*cur,animation[6] = {0,1,2,3,4,5};
 sprite_system *demo;
 buffer_rec *buff;
 unsigned int start_time,end_time; /* timing */
 list *dirty_recs;
 UBYTE *pal;
 drect *temp;
 list_node *save,*node;

 if(argc != 4){
    usage();
 }
 num_sprites=atoi(argv[1]);
 num_iterations=atoi(argv[2]);
 seed=atoi(argv[3]);

 if((num_sprites<=0)){
    usage();
 }
 if((num_iterations<=0)){
    usage();
 }

 /* allocate space for storing our rectangle array */
 dirty_recs=list_init();

 /* init stuff*/
 buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
 demo = sprite_init(num_sprites,6);
 pal = pal_load("spin.pal");
 
 if( sprite_load("spin.spr",demo) != SUCCESS ){
    puts("error: Couldn't load sprites");
    exit(-1);
 }
 
 if((buff== NULL)||(demo== NULL)||(pal== NULL)||(dirty_recs== NULL)){
    puts("Initialisation error.");
    exit(-1);
 }
 
 screen_set_video_mode();
 screen_block_set_pal(pal);
 srand(seed);
 uclock_init();  /* set up our timing */

 /* Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     drect_register(dirty_recs,x,y,x+SPR_MAX_X,y+SPR_MAX_Y);

     sprite_set_xy(demo,i,x,y);
     sprite_set_an_frame(demo,i,0);
     sprite_set_move_info(demo,i,0,0,0); /* not moving */
     sprite_set_anim_info(demo,i,1,6,animation);
     sprite_turn_on(demo,i);
 }

 fgetc(stdin);
 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);

    /* Draw each dirty rectangle to the screen */
    node=dirty_recs->head;
    while(node!=NULL){
       temp=(drect *)list_get_data_pointer(node);
       screen_blit_buff_toNC(temp->x1,temp->y1,buff,temp->x1,temp->y1,temp->x2,temp->y2);
       /* fgetc(stdin); */
       node=node->next;
    }

    buff_rest_all_sprites(demo,buff);
    sprite_update_all_anim_and_move(demo);
 }
 end_time=uclock_read();
 printf("%d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);
    screen_blit_fs_buffer(buff);
    buff_rest_all_sprites(demo,buff);
    sprite_update_all_anim_and_move(demo);
 }
 end_time=uclock_read();
 printf("fs buff %d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 screen_restore_video_mode();
 return 0; 
}

void usage(void)
{
 puts("spin3.exe : Speed up blitting with dirty rectangles blits.");
 puts("usage: spin3.exe <num_sprites> <num_iterations> <rand_seed>");
 puts("Output is on stdout, so pipe it to a file to view.");
 exit(0);
}

/* this function assumes x1<x2 and y1<y2 */
void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2)
{
 drect *temp,*lookup;
 list_node *node,*del;

 puts("this rect:");

 /* trivial rejection: any rectangles completely offscreen */
 if((x1>SCREEN_MAX_X)||(y1>SCREEN_MAX_Y)||(x2<0)||(y2<0)){
    puts("reject");
    return;
 }

 /* clip rectangle to the screen sized buffer */
 if(x2 > SCREEN_MAX_X) {
    puts("clipa");
    x2 = SCREEN_MAX_X;
 }
 if(y2 > SCREEN_MAX_Y) {
    puts("clipb");
    y2 = SCREEN_MAX_Y;
 }
 if(x1 <0){
    puts("clipc");
    x1 = 0;
 }
 if(y1 <0) {
    puts("clipd");
    y1 = 0;
 }
 
 temp=malloc(sizeof(drect));
 if(temp==NULL){
    jlib_exit(jlib_msg(JLIB_ENULL));
 }
 temp->x1=x1;
 temp->y1=y1;
 temp->x2=x2;
 temp->y2=y2;

 /* run through rectangles, picking up touching ones as we go */ 
 node=dirty_recs->head;
 while(node!=NULL){
    lookup=(drect *)list_get_data_pointer(node);
    if(drect_do_intersect(lookup,temp)){ /* adds lookup to temp if intersect */
       del=node;
       node=node->next;
       list_set_current(dirty_recs,del);
       list_remove_current(dirty_recs);
       puts("subsumed");
    }else{   
       node=node->next;
    }
 }
 puts("added");

 /* add to end of list */
 list_add_to_end(dirty_recs,(void *)temp);
}

#ifndef MIN
#define MIN(a,b)   (((a) < (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b)   (((a) < (b)) ? (b) : (a))
#endif


int drect_do_intersect(drect *one,drect *two)
{
 if( (one->x1 > two->x2)
   ||(one->x2 < two->x1)
   ||(one->y1 > two->y2)
   ||(one->y2 < two->y1)){
      return 0;
 }
 two->x1= MIN(one->x1,two->x1);\
 two->x2= MAX(one->x2,two->x2);\
 two->y1= MIN(one->y1,two->y1);\
 two->y2= MAX(one->y2,two->y2);
 return 1;
}

