/* Copyright 1995-96 Jon Griffiths.  See the file "jlib.doc" for details.
*
* This is an attempt at a much more dynamic dirty rectangle framework.
*
* Compares speed of dirty rectangles and blit_fs_buffer().
*
* Note: I'm using the list instructions to let me code this quickly.
*       A realistic implementation may be array based.
*
* The basic method:
* 1. Whenever an item in the offscreen buffer changes, we add the rectangle
*    that encompasses that change to a dirty rectangle list.
* 2. In this case we use a sprites x and y, and its frames width and height.
* 2. Rects that overlap are collapsed together avoiding flickering overlap.
* 3. As the dirty rectangles are drawn they are deleted from the list.
* 4. The list is then re-created for the next frame.
*
* Unbelievably, even this really unoptimised version improves speed for up
* to 25 64x64 sprites at 320x200 (which is 1/2 the screen worth of spr pixels).
* The speed improvement depends on how slow your video is, the slower your
* video, the more marked the improvement. The smaller or fewer the sprites
* to display, the better the algorithm performs.
*
* At present all sprites are added to the rectangle list even if they haven't
* moved or changed animation frame. This is wasteful- only changed sprites
* should be added (see bug note below).
*
* The rectangles are collapsed together in an inefficient way also, in that
* the resulting rectangle of collapsing A and B is a rectangle that fully
* encompasses them both, i.e:
*
* +---------+            +---------+         +---------+
* |   +-- -+|            |         |         |         |
* |   |    ||   Becomes: |         |  Where: |         |  Is Optimal.
* +---+----++            |         |         +---------+
*     |    |             |         |         XXXXX+---+X X=Wasted blitting,
*     |    |             |         |         XXXXX|   |X   as it hasn't
*     +----+             +---------+         XXXXX+---+X   changed.
*
* This should be fixed. Also, we should check the size of the list and go
* do a fs blit at some point that dirty rectangles won't save any time.
*
* NOTE: Requires JLib v1.6 to recompile.
*
* There is a bug with this approach: if a sprite moves say 5 pixels, then
*  we will only draw its new position rectangle, and will skip its old
*  position. Until another sprite moves over that part, the screen will
*  be inconsistent with the buffer. The solution for this is to either keep
*  two lists of rectangles or implement the dirty rectangle stuff into the
*  object movement code. The latter is preferred.
*/

#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include <jlib.h>

#define RANDOM(u) ( rand() % ( ((u) + 1)  ) )

void usage(void);
typedef struct {
        int x1,y1,x2,y2;
} drect;

void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2);
int drect_do_intersect(drect *one,drect *two);

int main(int argc,char *argv[])
{
 int num_iterations,num_sprites,i,j,seed,*cur,animation[6] = {0,1,2,3,4,5};
 sprite_system *demo;
 buffer_rec *buff;
 unsigned int start_time,end_time; /* timing */
 list *dirty_recs;
 UBYTE *pal;
 drect *temp;
 list_node *save,*node;

 if(argc != 4){
    usage();
 }
 num_sprites=atoi(argv[1]);
 num_iterations=atoi(argv[2]);
 seed=atoi(argv[3]);
 if((num_sprites<=0)){
    usage();
 }
 if((num_iterations<=0)){
    usage();
 }

 /* allocate space for storing our rectangle array */
 dirty_recs=list_init();

 /* init stuff*/
 buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
 demo = sprite_init(num_sprites,6);
 pal = pal_load("spin.pal");
 
 if( sprite_load("spin.spr",demo) != SUCCESS ){
    puts("error: Couldn't load sprites");
    exit(-1);
 }
 
 if((buff== NULL)||(demo== NULL)||(pal== NULL)||(dirty_recs== NULL)){
    puts("Initialisation error.");
    exit(-1);
 }
 
 screen_set_video_mode();
 screen_block_set_pal(pal);
 srand(seed);
 uclock_init();  /* set up our timing */

 /* Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     sprite_record *sr=demo->sprites[i];
     sprite_data_rec *sd;
     sd=demo->sprite_data[sr->frame];

     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     sprite_set_xy(demo,i,x,y);
     sprite_set_an_frame(demo,i,RANDOM(5));
     sprite_set_move_info(demo,i,RANDOM(2)+1,RANDOM(8)-4,RANDOM(8)-4);
     sprite_set_anim_info(demo,i,RANDOM(3)+1,6,animation);
     sprite_turn_on(demo,i);
     drect_register(dirty_recs,sr->x,sr->y,sr->x+sd->width+1,sr->y+sd->height+1);
 }

 fgetc(stdin);
 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);

    /* Draw each dirty rectangle to the screen, remove from list as we go */
    while(dirty_recs->head!=NULL){
       temp=(drect *)list_get_data_pointer(dirty_recs->head);
       screen_blit_buff_toNC(temp->x1,temp->y1,buff,temp->x1,temp->y1,temp->x2,temp->y2);
       list_remove_head(dirty_recs);
    }

    buff_rest_all_sprites(demo,buff);
    sprite_update_all_anim_and_move(demo);

    /* re-register the sprites in thier new positions/frames */
    for(j=0;j<num_sprites;j++){
        /* in a real world prog we would add these rects during our own
         * sprite updates. */
        sprite_record *sr=demo->sprites[j];
        sprite_data_rec *sd;
        sd=demo->sprite_data[sr->frame];
        drect_register(dirty_recs,sr->x,sr->y,sr->x+sd->width+1,sr->y+sd->height+1);
     }
 }
 end_time=uclock_read();
 printf("%d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 srand(seed);

 /* re-Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     sprite_set_xy(demo,i,x,y);
     sprite_set_an_frame(demo,i,RANDOM(5));
     sprite_set_move_info(demo,i,RANDOM(2)+1,RANDOM(8)-4,RANDOM(8)-4);
     sprite_set_anim_info(demo,i,RANDOM(3)+1,6,animation);
     sprite_turn_on(demo,i);
 }

 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);
    screen_blit_fs_buffer(buff);
    buff_rest_all_sprites(demo,buff);
    sprite_update_all_anim_and_move(demo);
 }
 end_time=uclock_read();
 printf("fs buff %d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 screen_restore_video_mode();
 return 0; 
}

void usage(void)
{
 puts("spin4.exe : Speed up blitting with dirty rectangles blits.");
 puts("usage: spin4.exe <num_sprites> <num_iterations> <rand_seed>");
 puts("Output is on stdout, so pipe it to a file to view.");
 exit(0);
}

/* This function assumes x1<x2 and y1<y2 and that we are drawing into a
 * fs buffer.
 */
void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2)
{
 drect *temp,*lookup;
 list_node *node,*del;

 /* trivial rejection: any rectangles completely offscreen */
 if((x1>SCREEN_MAX_X)||(y1>SCREEN_MAX_Y)||(x2<0)||(y2<0)){
    return;
 }

 /* clip rectangle to the screen sized buffer */
 if(x2 > SCREEN_MAX_X) {
    x2 = SCREEN_MAX_X;
 }
 if(y2 > SCREEN_MAX_Y) {
    y2 = SCREEN_MAX_Y;
 }
 if(x1 <0){
    x1 = 0;
 }
 if(y1 <0) {
    y1 = 0;
 }
 
 temp=malloc(sizeof(drect)); /* inefficient, array space would be better */

 if(temp==NULL){
    jlib_exit(jlib_msg(JLIB_ENULL));
 }

 temp->x1=x1;
 temp->y1=y1;
 temp->x2=x2;
 temp->y2=y2;

 /* run through rectangles, picking up touching ones as we go */ 
 /* not sure if we should add then join only once before drawing, or
  * join them together as we go, as is done here. */
 node=dirty_recs->head;
 while(node!=NULL){
    lookup=(drect *)list_get_data_pointer(node);
    if(drect_do_intersect(lookup,temp)){ /* adds lookup to temp if intersect */
       del=node;
       node=node->next;
       list_set_current(dirty_recs,del);
       list_remove_current(dirty_recs); /* we have subsumed rect, so delete it */
    }else{   
       node=node->next;
    }
 }
 /* add to end of list */
 list_add_to_end(dirty_recs,(void *)temp);
}

#ifndef MIN
#define MIN(a,b)   (((a) < (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b)   (((a) < (b)) ? (b) : (a))
#endif


int drect_do_intersect(drect *one,drect *two)
{
 /* should negate and use && to allow faster rejection here */
 if( (one->x1 > two->x2)
   ||(one->x2 < two->x1)
   ||(one->y1 > two->y2)
   ||(one->y2 < two->y1)){
      return 0;
 }
 /* if each case was spelled out, only some of this work would need to be done */
 two->x1= MIN(one->x1,two->x1);
 two->x2= MAX(one->x2,two->x2);
 two->y1= MIN(one->y1,two->y1);
 two->y2= MAX(one->y2,two->y2);
 return 1;
}

