/* Copyright 1995-96 Jon Griffiths.  See the file "jlib.doc" for details.
*
* This is an attempt at an integrated dirty rectangle framework.
*
* A dirty rectangle is added only if a sprite changes, this is done inside
* an adapted version of sprite_update_anim_and_move.
*
* Since the new and old position are updated together, this fixes the bug
* from spin4.c.
*
* Other optimisations have not been attempted.
*/

#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include <jlib.h>

#define RANDOM(u) ( rand() % ( ((u) + 1)  ) )

void usage(void);
typedef struct {
        int x1,y1,x2,y2;
} drect;

void drect_sprite_update_anim_and_move (sprite_system * spr_sys, int snum,list *alist);
void drect_sprite_update_all_anim_and_move (sprite_system * spr_sys,list *alist);
void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2);
int drect_do_intersect(drect *one,drect *two);

int main(int argc,char *argv[])
{
 int num_iterations,num_sprites,i,j,seed,*cur,animation[6] = {0,1,2,3,4,5};
 sprite_system *demo;
 buffer_rec *buff;
 unsigned int start_time,end_time; /* timing */
 list *dirty_recs;
 UBYTE *pal;
 drect *temp;
 list_node *save,*node;

 if(argc != 4){
    usage();
 }

 num_sprites=atoi(argv[1]);
 num_iterations=atoi(argv[2]);

 seed=atoi(argv[3]);
 if((num_sprites<=0)){
    usage();
 }
 if((num_iterations<=0)){
    usage();
 }

 /* allocate space for storing our rectangle array */
 dirty_recs=list_init();

 /* init stuff*/
 buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
 demo = sprite_init(num_sprites,6);
 pal = pal_load("spin.pal");
 
 if( sprite_load("spin.spr",demo) != SUCCESS ){
    puts("error: Couldn't load sprites");
    exit(-1);
 }
 
 if((buff== NULL)||(demo== NULL)||(pal== NULL)||(dirty_recs== NULL)){
    puts("Initialisation error.");
    exit(-1);
 }
 
 screen_set_video_mode();
 screen_block_set_pal(pal);
 srand(seed);
 uclock_init();  /* set up our timing */

 /* Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     sprite_record *sr=demo->sprites[i];
     sprite_data_rec *sd;
     sd=demo->sprite_data[sr->frame];

     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     sprite_set_xy(demo,i,x,y);
     sprite_set_an_frame(demo,i,RANDOM(5));
     sprite_set_move_info(demo,i,RANDOM(2)+1,RANDOM(8)-4,RANDOM(8)-4);
     sprite_set_anim_info(demo,i,RANDOM(3)+1,6,animation);
     sprite_turn_on(demo,i);
     drect_register(dirty_recs,sr->x,sr->y,sr->x+sd->width+1,sr->y+sd->height+1);
 }

 fgetc(stdin);
 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);

    /* Draw each dirty rectangle to the screen, remove from list as we go */
    while(dirty_recs->head!=NULL){
       temp=(drect *)list_get_data_pointer(dirty_recs->head);
       screen_blit_buff_toNC(temp->x1,temp->y1,buff,temp->x1,temp->y1,temp->x2,temp->y2);
       list_remove_head(dirty_recs);
    }

    buff_rest_all_sprites(demo,buff);
    drect_sprite_update_all_anim_and_move(demo,dirty_recs);
 }
 end_time=uclock_read();
 printf("%d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 srand(seed);

 /* re-Set up random sprite positions and animation */
 for(i=0;i<num_sprites;i++){
     int x,y;
     x=RANDOM(SCREEN_WIDTH);
     y=RANDOM(SCREEN_HEIGHT);

     sprite_set_xy(demo,i,x,y);
     sprite_set_an_frame(demo,i,RANDOM(5));
     sprite_set_move_info(demo,i,RANDOM(2)+1,RANDOM(8)-4,RANDOM(8)-4);
     sprite_set_anim_info(demo,i,RANDOM(3)+1,6,animation);
     sprite_turn_on(demo,i);
 }

 start_time=uclock_read();

 for(i=0;i<num_iterations;i++){
    buff_save_all_sprites(demo,buff);
    buff_draw_all_sprites(demo,buff);
    screen_blit_fs_buffer(buff);
    buff_rest_all_sprites(demo,buff);
    sprite_update_all_anim_and_move(demo);
 }
 end_time=uclock_read();
 printf("fs buff %d Iterations in %d usec.\n",num_iterations,end_time-start_time);

 screen_restore_video_mode();
 return 0; 
}

void usage(void)
{
 puts("spin5.exe : Speed up blitting with dirty rectangles blits.");
 puts("usage: spin4.exe <num_sprites> <num_iterations> <rand_seed>");
 puts("Output is on stdout, so pipe it to a file to view.");
 exit(0);
}

/* This function assumes x1<x2 and y1<y2 and that we are drawing into a
 * fs buffer.
 */
void drect_register(list *dirty_recs,int x1,int y1,int x2,int y2)
{
 drect *temp,*lookup;
 list_node *node,*del;

 /* trivial rejection: any rectangles completely offscreen */
 if((x1>SCREEN_MAX_X)||(y1>SCREEN_MAX_Y)||(x2<0)||(y2<0)){
    return;
 }

 /* clip rectangle to the screen sized buffer */
 if(x2 > SCREEN_MAX_X) {
    x2 = SCREEN_MAX_X;
 }
 if(y2 > SCREEN_MAX_Y) {
    y2 = SCREEN_MAX_Y;
 }
 if(x1 <0){
    x1 = 0;
 }
 if(y1 <0) {
    y1 = 0;
 }
 
 temp=malloc(sizeof(drect)); /* inefficient, array space would be better */

 if(temp==NULL){
    jlib_exit(jlib_msg(JLIB_ENULL));
 }

 temp->x1=x1;
 temp->y1=y1;
 temp->x2=x2;
 temp->y2=y2;

 /* run through rectangles, picking up touching ones as we go */ 
 /* not sure if we should add then join only once before drawing, or
  * join them together as we go, as is done here. */
 node=dirty_recs->head;
 while(node!=NULL){
    lookup=(drect *)list_get_data_pointer(node);
    if(drect_do_intersect(lookup,temp)){ /* adds lookup to temp if intersect */
       del=node;
       node=node->next;
       list_set_current(dirty_recs,del);
       list_remove_current(dirty_recs); /* we have subsumed rect, so delete it */
    }else{   
       node=node->next;
    }
 }
 /* add to end of list */
 list_add_to_end(dirty_recs,(void *)temp);
}

#ifndef MIN
#define MIN(a,b)   (((a) < (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b)   (((a) < (b)) ? (b) : (a))
#endif


int drect_do_intersect(drect *one,drect *two)
{
 /* should negate and use && to allow faster rejection here */
 if( (one->x1 > two->x2)
   ||(one->x2 < two->x1)
   ||(one->y1 > two->y2)
   ||(one->y2 < two->y1)){
      return 0;
 }
 /* if each case was spelled out, only some of this work would need to be done */
 two->x1= MIN(one->x1,two->x1);
 two->x2= MAX(one->x2,two->x2);
 two->y1= MIN(one->y1,two->y1);
 two->y2= MAX(one->y2,two->y2);
 return 1;
}

void drect_sprite_update_anim_and_move (sprite_system * spr_sys, int snum,list *alist)
{
   int changed_flag=0; 
   sprite_record *current;
   drect start,end;
   current = spr_sys->sprites[snum];

   if(SPR_IS_ON (spr_sys, snum)){
      /* record current rectangle */
      start.x1=current->x;
      start.y1=current->y;
      start.x2=current->x+spr_sys->sprite_data[current->frame]->width;
      start.y2=current->y+spr_sys->sprite_data[current->frame]->height;

      if(current->mode == SPR_MODE_WHOLE){
	 /* check movement */
	 if (current->speed){
	      if ((--current->speed_counter) == 0){
                   changed_flag=1;
                   current->speed_counter = current->speed;
		   current->x += current->xinc;
		   current->y += current->yinc;
              }
         }

	 /* check animation */
	 if (current->animspeed){
	      if ((--current->animspeed_counter) == 0){
                   changed_flag=1;
                   current->animspeed_counter = current->animspeed;

		   if ((++current->frame_count) == current->noframes){
		       current->frame_count = 0;
	           }
        	   current->frame = current->animframes[current->frame_count];
              }
          }
    }
    else{
	 /* check fixed point movement */
	 if (current->speed){
             changed_flag=1;
	     current->speed_counter+=current->speed;
	     current->x += ((current->speed_counter>>4)*current->xinc);
	     current->y += ((current->speed_counter>>4)*current->yinc);
	     current->speed_counter=(current->speed_counter & 0xf);
	 }

	 /* check fixed point animation */
	 if (current->animspeed){
             changed_flag=1;
             current->animspeed_counter+=current->animspeed;

	     current->frame_count+=(current->animspeed_counter>>4);
	     
	     while(current->frame_count >= current->noframes){
		current->frame_count -= current->noframes;
	     }
	     
	     current->frame = current->animframes[current->frame_count];
	 }
	 current->animspeed_counter=(current->animspeed_counter & 0xf);
    }     
         if(changed_flag){
           end.x1=current->x;
           end.y1=current->y;
           end.x2=current->x+spr_sys->sprite_data[current->frame]->width;
           end.y2=current->y+spr_sys->sprite_data[current->frame]->height;
           drect_register(alist,start.x1,start.y1,start.x2,start.y2);
           drect_register(alist,end.x1,end.y1,end.x2,end.y2);
         }
       }
}


void drect_sprite_update_all_anim_and_move (sprite_system * spr_sys,list *alist)
{
   int count;

   for (count = 0; count < spr_sys->no_sprites; count++){
	if (SPR_IS_ON (spr_sys, count)){
             drect_sprite_update_anim_and_move (spr_sys, count,alist);
	}
   }
}

