/* Copyright 1996  Jonathan Griffiths. See "readme.jon" for detailed terms. */
#include "level.h"

/*+------------------------------------------------------------------------+*/
/*| Create a blank level of the given size.                                |*/
/*+------------------------------------------------------------------------+*/
level *level_init(int x,int y,sprite_system *sys)
{
 level *newlevel;

 JLIB_ENTER("level_init");

 /* malloc space for record */
 if((newlevel=(level *)malloc(sizeof(level)))==NULL){
    jlib_exit(jlib_msg(JLIB_EMALLOC));
 }

 if(sys==NULL){
    jlib_exit(jlib_msg(JLIB_ENULL));
 }

 /* set level details */
 newlevel->magic=LEVEL_MAGIC;

 newlevel->width=x;
 newlevel->height=y;
 newlevel->block_width=sys->sprite_data[0]->width;
 newlevel->block_height=sys->sprite_data[0]->height;
 newlevel->num_blocks=SPR_NUM_LOADED(sys);

 if((newlevel->map=(UBYTE *)malloc(newlevel->width*newlevel->height))==NULL){
    jlib_exit(jlib_msg(JLIB_EMALLOC));
 }

 /* blank the memory */
 memset(newlevel->map,0,newlevel->width*newlevel->height);

 newlevel->blocks=sys;

 JLIB_LEAVE;
 return newlevel;
}

/*+------------------------------------------------------------------------+*/
/*| Load a level into memory.                                              |*/
/*+------------------------------------------------------------------------+*/
level *level_load(char *filename,sprite_system *sys)
{
 int i;
 level *newlevel;
 FILE *fp;

 JLIB_ENTER("level_load");

 /* malloc space for record */
 if((newlevel=(level *)malloc(sizeof(level)))==NULL){
    jlib_exit(jlib_msg(JLIB_EMALLOC));
 }

 /* open level file */
 if((fp=fopen(filename,"rb"))== NULL){
        JLIB_SPRINTF("Couldn't open file %s.", filename);
        free(newlevel);
        JLIB_LEAVE;
        return NULL;
 }

 /* load level details */
 newlevel->magic=fgetc(fp);

 if(newlevel->magic!=LEVEL_MAGIC){
    JLIB_SPRINTF("File %s not a level file.",filename);
    free(newlevel);
    JLIB_LEAVE;
    return NULL;
 }

 newlevel->width=fgetc(fp);
 newlevel->height=fgetc(fp);
 newlevel->block_width=sys->sprite_data[0]->width;
 newlevel->block_height=sys->sprite_data[0]->height;

 newlevel->num_blocks=SPR_NUM_LOADED(sys);

 if((newlevel->map=(UBYTE *)malloc(newlevel->width*newlevel->height))==NULL){
    jlib_exit(jlib_msg(JLIB_EMALLOC));
 }

 /* load in level map */
 for(i=0;i<(newlevel->width*newlevel->height);i++){
    newlevel->map[i]=fgetc(fp);
 }

 newlevel->blocks=sys;
 fclose(fp);

 JLIB_LEAVE;
 return newlevel;
}


/*+------------------------------------------------------------------------+*/
/*| Save a level to a file.                                                |*/
/*+------------------------------------------------------------------------+*/
int level_save(char *filename,level *lev)
{
 int i;
 FILE *fp;

 JLIB_ENTER("level_save");

 /* open level file */
 if((fp=fopen(filename,"wb"))== NULL){
        JLIB_SPRINTF("Couldn't open file %s.", filename);
        free(lev);
        JLIB_LEAVE;
        return 0;
 }

 fputc(lev->magic,fp);
 fputc(lev->width,fp);
 fputc(lev->height,fp);

 /* load in level map */
 for(i=0;i<(lev->width*lev->height);i++){
    fputc(lev->map[i],fp);
 }
 fclose(fp);

 JLIB_LEAVE;
 return 1;
}


/*+------------------------------------------------------------------------+*/
/*| Free the memory used by a level.                                        |*/
/*+------------------------------------------------------------------------+*/
level *level_free(level *level)
{
 JLIB_ENTER("level_free");

 if(level!=NULL){
   if(level->blocks!=NULL){
     sprite_free(level->blocks);
   }
   free(level);
 }

 JLIB_LEAVE;
 return NULL;
}


/*+------------------------------------------------------------------------+*/
/*| Stencil a level into a buffer with clipping.                           |*/
/*+------------------------------------------------------------------------+*/
void level_stencil(level *lev,buffer_rec *buff,int x,int y)
{
 int xpos,ypos,xoff,yoff;
 int w,h,b;
 UBYTE *map;

 if(x){
    xpos=x/lev->block_width;
    xoff=x%lev->block_width;
 }
 else{
    xpos=0;
    xoff= lev->block_width;
 }

 if(y){
    ypos=y/lev->block_height;
    yoff=y%lev->block_height;
 }
 else{
    ypos=0;
    yoff= lev->block_height;
 }

 w=(B_X_SIZE(buff)/lev->block_width)+2;
 h=(B_Y_SIZE(buff)/lev->block_height)+2;

 /* handle trivial clipping cases */
 if((xpos>lev->width)||(ypos>lev->height)||(xpos+w<0)||(ypos+h<0)){
    return;
 }

 /* handle other clipping cases */
 if((xpos+w)>lev->width){
    w=lev->width-xpos;
 }

 if(xpos<0){
    w+=xpos;
    x=((-xpos)*lev->block_width)-xoff;
    xpos= 0;
 }
 else{
    if(x!=0){
      x= -xoff;
    }
 }


 if((ypos+h)>lev->height){
    h=lev->height-ypos;
 }

 if(ypos<0){
    h+=ypos;
    y=((-ypos)*lev->block_height)-yoff;
    ypos= 0;
 }
 else{
    if(y!=0){
      y= -yoff;
    }
 }

 map=lev->map+ypos*lev->width+xpos;
 
 while(h--){
    int nx=x;

    for(b=w;b!=0;b--){
       if(*map){
          buff_stencil_sprite(lev->blocks,*map,buff,nx,y);
       }
       nx+=lev->block_width;
       map++;
    }
    map+=(lev->width-w);
    y+=lev->block_height;
 }
}

 
/*+------------------------------------------------------------------------+*/
/*| Stamp a level into a buffer with clipping.                             |*/
/*+------------------------------------------------------------------------+*/
void level_stamp(level *lev,buffer_rec *buff,int x,int y)
{
 int xpos,ypos,xoff,yoff;
 int w,h,b;
 UBYTE *map;

 if(x){
    xpos=x/lev->block_width;
    xoff=x%lev->block_width;
 }
 else{
    xpos=0;
    xoff= lev->block_width;
 }

 if(y){
    ypos=y/lev->block_height;
    yoff=y%lev->block_height;
 }
 else{
    ypos=0;
    yoff= lev->block_height;
 }

 w=(B_X_SIZE(buff)/lev->block_width)+2;
 h=(B_Y_SIZE(buff)/lev->block_height)+2;

 /* handle trivial clipping cases */
 if((xpos>lev->width)||(ypos>lev->height)||(xpos+w<0)||(ypos+h<0)){
    return;
 }

 /* handle other clipping cases */
 if((xpos+w)>lev->width){
    w=lev->width-xpos;
 }

 if(xpos<0){
    w+=xpos;
    x=((-xpos)*lev->block_width)-xoff;
    xpos= 0;
 }
 else{
    if(x!=0){
      x= -xoff;
    }
 }


 if((ypos+h)>lev->height){
    h=lev->height-ypos;
 }

 if(ypos<0){
    h+=ypos;
    y=((-ypos)*lev->block_height)-yoff;
    ypos= 0;
 }
 else{
    if(y!=0){
      y= -yoff;
    }
 }

 map=lev->map+ypos*lev->width+xpos;
 
 while(h--){
    int nx=x;

    for(b=w;b!=0;b--){
       if(*map){
          buff_stamp_sprite(lev->blocks,*map,buff,nx,y);
       }
       nx+=lev->block_width;
       map++;
    }
    map+=(lev->width-w);
    y+=lev->block_height;
 }
}
