/* Copyright 1996  Jonathan Griffiths. See "readme.jon" for detailed terms. */
#include "list.h"

list_node *init_node(void);
void free_node(list_node *node);


/* Set up the list ------------------------------------------------------- */
list *list_init(void)
{
 list *new;

 JLIB_ENTER("list_init");

 new = (list *)malloc(sizeof(list));

 new->head = new->tail = new->current = (list_node *)NULL;

 JLIB_LEAVE;
 return(new);
}


/* Destroy the list ------------------------------------------------------ */
list *list_destroy(list *listp)
{
 JLIB_ENTER("list_destroy");

 while(listp->head != NULL){
    list_remove_head(listp);
 }

 free((void *)listp);

 JLIB_LEAVE;
 return (list *)NULL;
}


/* Add an item to the start of the list ---------------------------------- */
void list_add_to_start(list *list,void *list_item)
{
 list_node *node;

 JLIB_ENTER("list_add_to_start");

 node=init_node();

 node->data=list_item;
 node->prev=NULL;
 list->current=node;

 if(list->head == NULL) {
    list->head = list->tail = node;
    node->next = NULL;
  }
  else{
       list->head->prev = node;
       node->next = list->head;
       list->head = node;
 }

 JLIB_LEAVE;
}


/* Add an item to the end of the list ------------------------------------- */
void list_add_to_end(list *list,void *list_item)
{
 list_node *node;

 JLIB_ENTER("list_add_to_end");

 node = init_node();
 node->data = list_item;
 node->next = NULL;
 list->current = node;

 if(list->tail == NULL) {
    list->tail = list->head = node;
    node->prev = NULL;
 }
 else{
    list->tail->next = node;
    node->prev = list->tail;
    list->tail = node;
 }

 JLIB_LEAVE;
}


/* Add an item after the current item in the list ------------------------ */
void list_add_after_current(list *list,void *list_item)
{
  list_node *node;

  JLIB_ENTER("list_add_after_current");

  node = init_node();
  node->data = list_item;

  if(list->current == NULL){
     list->current=list->head;
  }

  if(list->head == NULL) {
     list->tail = list->head = list->current = node;
     node->prev = node->next = NULL;
  }
  else{
      if(list->current->next == NULL) {
	 list->current->next = node;
	 node->next = NULL;
	 node->prev = list->current;
         list->tail = list->current = node;
      }
      else{
	  node->prev = list->current;
	  node->next = list->current->next;
          node->next->prev = node;
	  list->current->next = node;
	  list->current = node;
      }
  }

 JLIB_LEAVE;
}


/* add an item before the current item in the list ----------------------- */
void list_add_before_current(list *list,void *list_item)
{
  list_node *node;

  JLIB_ENTER("list_add_before_current");

  node = init_node();
  node->data = list_item;

  if(list->current == NULL){
     list->current=list->head;
  }
  if(list->head == NULL) {
     list->tail = list->head = list->current = node;
     node->prev = node->next = NULL;
  }
  else{
      if(list->current->prev == NULL) {
	 list->current->prev = node;
	 node->prev = NULL;
	 node->next = list->current;
         list->head = list->current = node;
      }
      else{
	  node->next = list->current;
	  node->prev = list->current->prev;
	  node->prev->next = node;
          list->current->prev = list->current = node;
      }
  }

 JLIB_LEAVE;
}


/* Remove the current item in the list ----------------------------------- */
void list_remove_current(list *list)
{
  list_node *temp;

  JLIB_ENTER("list_remove_current");

  if(list->current == NULL){
     JLIB_LEAVE;
     return;
  }

  if(list->current == list->head){
     list_remove_head(list);
  }
  else{
     if(list->current == list->tail){
        list_remove_tail(list);
     }
     else{
	 temp = list->current;
         temp->prev->next = temp->next;
         temp->next->prev = temp->prev;
	 list->current = temp->next;
         free_node(temp);
     }
  }

  JLIB_LEAVE;
}


/* Remove the head of the list ------------------------------------------- */
void list_remove_head(list *list)
{
  list_node *temp;

  JLIB_ENTER("list_remove_head");

  if(list->head == NULL){
     JLIB_LEAVE;
     return;
  }

  if(list->head == list->tail){
     temp = list->head;
     list->head = list->tail = list->current = NULL;
  }
  else{
     temp = list->head;
     list->head = temp->next;
     list->head->prev = NULL;
     list->current = list->head;
  }

  free_node(temp);

  JLIB_LEAVE;
}

/* remove the tail of the list ------------------------------------------- */
void list_remove_tail(list *list)
{
  list_node *temp;

  JLIB_ENTER("list_remove_tail");

  if(list->tail == NULL){
     JLIB_LEAVE;
     return;
  }

  if(list->tail == list->head){
     temp = list->tail;
     list->head = list->tail = list->current = NULL;
  }
  else{
     temp = list->tail;
     list->tail = temp->prev;
     list->tail->next = NULL;
     list->current = list->tail;
  }

  free_node(temp);

  JLIB_LEAVE;
}


/* Create a new node in the list ----------------------------------------- */
list_node *init_node(void)
{
 list_node *new;

 JLIB_ENTER("init_node");

 new = (list_node *)malloc(sizeof(list_node));

 if(new==NULL){
    jlib_exit("malloc failure");
 }

 JLIB_LEAVE;
 return new;
}


/* free a list node and its data ----------------------------------------- */
void free_node(list_node *node)
{
 JLIB_ENTER("free_node");

 free((void *)node->data);
 free((void *)node);

 JLIB_LEAVE;
}
