
  /* /////////////////////////////////////////////////////////////// */
 /* /// BMP_CONV.C                                              /// */
/* /////////////////////////////////////////////////////////////// */
	/* --- Elmar Vogt - V1.0 - 1.96 - DJGPP C 2.003 --------- */

/* this file contains the source code for our conversion utilities,
	which allow transformation between windows bmp format (as used
	by paintbrush and several other commercial programs), and xlib
	pbm formats */

/* *************************************************************** */
/* *** IMPORTANT NOTE :	this software comes without warranty 	*** */
/* *** of any kind; it is public domain and not for resale.		*** */
/* *** copyright exists in portions of this software.				*** */
/* *** please, refer to the file BMP_CONV.DOC for details.		*** */
/* *************************************************************** */

/* === include files =============================================== */
#include "bmp_conv.h"

/* === global variable and constant definitions ==================== */
/* _bmp_header_ displays the standard format for a bmp-file start
	(windows 3.x). some values can change from image to image, but
	most of them are constant for a given mode.
	the comments at the beginning of every line give the offset from
	the file start, and whether the values have to be considered a
	4 byte long ('l'), or a 2 byte int ('i'). */
static BYTE bmp_header[54]={
		/* a) file header section */
	/*  0 - */	66, 77,		  		/* 'BM', file type id */
	/*  2 l */	0, 0, 0, 0,	  		/* total file length */
	/*  4 i */	0, 0,			  		/* reserved */
	/*  8 i */	0, 0,			  		/* reserved */
	/* 10 l */	54, 4, 0, 0,  		/* offset for image data, 54 header
												bytes + 4*256 palette bytes, if
												palette is used */
		/* b) image information section */
	/* 14 l */	40, 0, 0, 0,  		/* information section length (?) */
	/* 18 l */	104, 1, 0, 0, 		/* image width/[pixel], here: 360 */
	/* 22 l */	224, 1, 0, 0, 		/* image height/[pixel], here: 480 */
						  	/* preset values for maximum resolution obtained
					  			with xlib */
	/* 26 i */	1, 0,		  			/* color planes, always 1 */
	/* 28 i */	8, 0,			  		/* bit/pixel color coding: one of
												1, 8, or 256 */
	/* 30 l */	0, 0, 0, 0,	  		/* compression scheme (none mostly) */
	/* 34 l */	0, 0, 0, 0,	  		/* image data section length */
	/* 38 l */	0, 0, 0, 0,	  		/* resolution pixel/meter in x-dir. */
	/* 42 l */	0, 0, 0, 0,	  		/* resolution pixel/meter in y-dir. */
					  		/* preset to '0' */
	/* 46 l */	0, 1, 0, 0,	  		/* number of colors used, 0="all" */
	/* 50 l */	0, 1, 0, 0	  		/* no. of important colors, 0="all" */
		/* if color coding equals 8, here follows the color palette:
			4 bytes per color, arrange: blue, green, red, (reserved);
			256 colors are stored, requiring 1024 byte. */
		/* now follows the image data block: if color coding is 8, it's
			one byte/pixel, else 3 byte/pixel.
			they are arranged in rows, each row _image width_ pixel wide.
			the order of rows is bottom line first, top line last. (sic!)
			this requires some unwieldy conversions on our part. */
	};

/* the following constants denote offset indices for the start of
	several important data in the bmp file header: */
static const int 	BMP_LENGTH= 			2,
						BMP_DATA_OFFSET=		10,
						BMP_WIDTH=				18,
						BMP_HEIGHT=				22,
						BMP_COLORS=				28,
						BMP_DATA_LENGTH=		34,
						BMP_PALETTE_START=	54;
		/* capitalization used to indicate use as constant index! */

/* note:  things could be make considerably easier, if one set up a
	struct that holds the bmp data header. one could simply make a struct
	pointer point towards the address in question. unfortunately, due to
	the unpredictable behaviour of some compilers, they might perform
	substantial data alignment in a struct, which would lead to a
	particular mismatch with the bmp data structure. */

/* these are the different color encoding modes: */
enum{BMP_BW=1, BMP_16= 4, BMP_256=8, BMP_TRUE_COLOR= 24};
		/* color encoding modes: BW, 256 colors w/ palette, and true
			color, resp.
			they are enumerated, since the switch/case-construct doesn't
			recognize const int's as integer constants. (yes, try it some
			day!) */

/* if the palette is not to be overwritten, we require a LUT to do
	a palette transformation: */
static BYTE palette_conversion_LUT[256];

/* === conversion routines ========================================= */

/* --- load_bmp(BYTE **, BYTE *, char *, BOOLEAN, CORRECT_DISTORTION) */
/* this function is used to load a bmp-formatted image file named
	_bmp_source_name_ from disk and convert it to a planar bitmap
	in the xlib-style. target is _pbm_target_, essentially a pointer
	to an array of BYTEs. no pre-allocation of memory is necessary.

	the conversion is done by setting up an intermediate linear bitmap,
	which in turn is converted to a planar bitmap.

	unless we have to deal with a true-color image, this function is
	pretty fast, and the speed limit is set by the disk drive. true-
	color images, on the other hand, may require processing times of
	up to one minute with a 486@33MHz!

	black and white images are not yet implemented.

	_palette_ is the color palette employed. if _palette_overwrite_ is
	_TRUE_, the colors from the loaded image will replace the colors
	used before. you may not want this, especially if you'd like to
	view several images at the same time. therefore, if _palette_over-
	write_ is _FALSE_, no overwrite will take place, but the colors
	of the loaded image will be approximated by a least square method
	to the colors of the current palette. the current palette will not
	be influenced in this case.
	even if _palette_overwrite_ is employed, you will not see any
	effects until you claim an update of the palette with
	_x_set_palette()_.
	for true color images, the palette approximation is always done.

	return value is the length of the PBM in bytes, '0' in case of an
	error. */

int load_bmp(BYTE **pbm_target, BYTE *palette, char *bmp_source_name,
	BOOLEAN palette_overwrite, CORRECT_DISTORTION distortion_mode){

	BYTE 	*bmp_data,			/* buffer for the loaded file */
			*linear_data;		/* buffer for the linear bitmap */
	long	bmp_file_length,	/* data obtained from the bmp- */
			bmp_data_length,	/* file header */
			bmp_data_offset,
			pbm_length,
			linear_index,		/* indices for converision of */
			bmp_index,			/* bmp to linear bitmap */
			palette_index;
	int	bmp_mode;			/* color encoding mode */

	int	x, y,					/* loop var's */
			bmp_x_ext, x_ext, x_ext2, y_ext,
									/* width and height of the picture.
										_x_ext_ is necessary, since pbm's have
										to be a multiple of 4 pixel wide. */
			line_step, pixel_step;
									/* steps utilized when converting a true
										color image */

	BYTE	red1, red2, green1, green2, blue1, blue2;
									/* color components for correcting an anti-
										aliased image */

	BOOLEAN
			dNONE=		(distortion_mode==NONE),
			dSHRINK=		(distortion_mode==SHRINK),
			dMAGNIFY=	(distortion_mode==MAGNIFY),
			dANTIALIAS=	(distortion_mode==ANTIALIAS);

	/* first, estimate required space, then allocate buffer memory: */
	bmp_file_length= file_sizeof(bmp_source_name);
		/* in the doc's of xlib v2.5, _file_sizeof()_ is still called
			_x_sizeof()_ */
	bmp_data= byte_array(bmp_file_length);

	/* read in the file data: */
	read_binary(bmp_source_name, &bmp_data);

	/* read the image parameters from the file buffer: */
	bmp_data_length= read_long_intel(&bmp_data[BMP_DATA_LENGTH]);
	bmp_data_offset= read_int_intel(&bmp_data[BMP_DATA_OFFSET]);
		/* actually a 4 byte long, but 4th byte seems corrupt ??? */

	/* find out the color resolution mode employed: */
	bmp_mode= read_int_intel(&bmp_data[BMP_COLORS]);

	/* get image dimensions: */
	bmp_x_ext= (x_ext= read_int_intel(&bmp_data[BMP_WIDTH]));
	y_ext= read_int_intel(&bmp_data[BMP_HEIGHT]);
		/* actually, these are longs, but xlib can handle int's for
			coordinates only. anyway, who has a monitor with >64k
			pixels per row...? */

	/* modify pbm dimensions according to distortion correction mode
		used: */
	if(dNONE){
			/* dimensions are correct */
	}
	else if(dSHRINK){
		x_ext/= 2;
	}
	else if(dMAGNIFY || dANTIALIAS){
		/* double number of rows */
		y_ext*= 2;
		if(dANTIALIAS)
			y_ext--;
			/* if antialiasing is engaged, we need an	odd number of lines,
				with the last line belonging to the original picture (not
				interpolated) */
	}
	else{
		x_set_text_mode();
		fprintf(stderr, "Mode for distortion correction %i could not be\
			recognized. \n", distortion_mode);

		return(0);
			/* abort mission */
	}

	/* prepare extension in x-direction to be a multiple of 4;
		necessary for conversion BM -> PBM: */
	x_ext2= x_ext;
	x_ext= align_long(x_ext);

	/* assign memory for intermediate buffer (_linear_data_), and
		target PBM: */
	pbm_length= 4 + (x_ext*y_ext);
	linear_data= byte_array(pbm_length);
	*pbm_target= byte_array(pbm_length);

	/* set width and height in linear bitmap: */
	write_int_intel(&linear_data[0], x_ext);
	write_int_intel(&linear_data[2], y_ext);
		/* transfer information in two-byte format! */

	switch(bmp_mode){

		case (BMP_256):

			/* means we have a 256-color image with palette */
			bmp_index= bmp_data_offset;
				/* though image width is free, the actual data rows have
					to be stuffed with zeroes to provide LONG alignment! */
			line_step= 1 + (dMAGNIFY || dANTIALIAS);
				/* jump one pbm line, only if _MAGNIFY_ or _ANTIALIAS_ */

			for(y=y_ext-dMAGNIFY-1; y>=0; y-= line_step){
				/* row by row, bottom up! */
				linear_index= 4+x_ext*y;

				if(dSHRINK){
					for(x=0; x<x_ext2; x++){
						linear_data[linear_index++]= bmp_data[bmp_index];
						bmp_index+= 2;
					}
					if(bmp_x_ext%4)
						bmp_index= align_long(++bmp_index);
				}
				else{
					memcpy(&linear_data[linear_index], &bmp_data[bmp_index],
						x_ext);

					if(dMAGNIFY){
						/* copy the line one row below */
						memcpy(&linear_data[linear_index+x_ext],
							&linear_data[linear_index], x_ext);
					}

					bmp_index= align_long(bmp_index+bmp_x_ext);
				}
			}

			/* palette adaption, as necessary: */
			if(palette_overwrite){
				palette_make_pbm(palette, &bmp_data[BMP_PALETTE_START]);
			}
			else{
				make_best_palette_match(palette_conversion_LUT, palette,
					&bmp_data[BMP_PALETTE_START]);
				/* in this case, we have to correct all the pixel values
					of _linear_data_ by means of the palette LUT: */
				for(linear_index=4; linear_index<pbm_length;
					linear_index++){

					linear_data[linear_index]=
						palette_conversion_LUT[linear_data[linear_index]];
				}
			}

			break;

		case (BMP_TRUE_COLOR):

			/* it's a true color image which we have to convert pixel by
				pixel always- this can take some time. */

			pixel_step= 3 * (1+dSHRINK);
				/* if _SHRINK_, then jump one bmp pixel */
			line_step= 1 + (dMAGNIFY || dANTIALIAS);
				/* jump one pbm line, only if _MAGNIFY_ or _ANTIALIAS_ */

			bmp_index= bmp_data_offset;
			for(y=y_ext-1; y>=0; y-= line_step){
				/* _y_ext_ is the number of lines in the target PBM! */
				linear_index= 4 + (y*x_ext);

				for(x=0; x<bmp_x_ext; x+= 1+dSHRINK){
					linear_data[linear_index++]=
						best_color_match(palette, bmp_data[bmp_index+2],
						bmp_data[bmp_index+1], bmp_data[bmp_index]);

					bmp_index+= pixel_step;
 					/* x_put_pixel(x, y, Page0_Offs,
 					linear_data[linear_index-1]); */
					/* please, remove above commment brackets to watch	the
						color conversion progressing! */
				}

				/* check _bmp_index_: */
				if(dSHRINK)
					bmp_index-= (bmp_x_ext%2)*3;
					/* if the number of pixels is odd, we have to move one
						pixel=3 bytes backwards */
				bmp_index+= (bmp_x_ext%4);
					/* correct index for long-alignment */
			}

			if(dMAGNIFY){
				for(y=0; y<y_ext; y+=2){
						/* copy screen line: */
					linear_index=4+ (y*x_ext);
					memcpy(&linear_data[linear_index+x_ext],
						&linear_data[linear_index], x_ext);
				}
			}

			break;

		case (BMP_16):
			/* 16 fixed colors */
		case (BMP_BW):
			/* it's just black and white */
			x_set_text_mode();
			fprintf(stderr, "File <%s> is a Black-and-White or 16 color \
				bitmap image.\n",	bmp_source_name);
			fprintf(stderr, "This mode is not yet implemented.\n");

			pbm_length= 0;
			break;

		default:
			/* unknown mode */
			x_set_text_mode();
			fprintf(stderr, "File <%s> does not belong to any identifiable\
				color \nmapping scheme.  Mode:  %i.  \n", bmp_source_name,
				bmp_mode);

			pbm_length= 0;
			break;
  	}

	/* after all the rest, fill the gaps for the antialiased
		image: */
	if(pbm_length && dANTIALIAS){
		for(y=1; y<y_ext; y+=2){
			linear_index= 4+x_ext*y;

			for(x=0; x<x_ext; x++){

				palette_index= linear_data[linear_index-x_ext]*3;
					/* this is the palette index which indicates the
						color used in the pixel one row above the one
						that's to be calculated */
				red1=		palette[palette_index++];
				green1=	palette[palette_index++];
				blue1=	palette[palette_index];

				palette_index= linear_data[linear_index+x_ext]*3;
					/* the same, one row below */
				red2=		palette[palette_index++];
				green2=	palette[palette_index++];
				blue2=	palette[palette_index];

				linear_data[linear_index++]=
					best_color_match(palette, (red1+red2)*2,
					(green1+green2)*2, (blue1+blue2)*2);
					/* to find the mean, usually you would have to use
						e.g.  (red1+red2)/2. but _best_color_mach()_
						assumes its parameters to come from a palette
						with maximum intensity=255. since we have only
						63 for a maximum value, we have to correct them
						by a factor of 4, or all in all multiply by 2.
						*/
				/* x_put_pixel(x, y, Page0_Offs,
				linear_data[linear_index-1]); */
			}
		}
	}

	/* final converison step: linear bitmap -> planar bitmap;
		nicely done by Paul's xlib: */
	if(pbm_length)
		x_BM_to_PBM(linear_data, *pbm_target);

	/* never leave a mess behind: */
	free(linear_data);
	free(bmp_data);

	return pbm_length;
}

/* --- save_bmp_box(char *, int, int, int, int, BYTE *) ------------ */
/* this function does NOT save a planar bitmap, but takes a snapshot
	directly from the screen. reasons for this are (a) you usually
	will not manipulate a bitmap directly, but rather save some screen
	composition, (b) if really necessary, you can always clear the
	screen, put your bitmap on it and save the screen region (though
	this is not very elegant, I admit :-), and (c) i have no idea about
	the planar bitmap format.

	to save on writing, the screen region is always captured from the
	screen with page offset _Page0_Offs_. (personally, i almost always
	use high res with a single screen only anyway.)
	the screen snapshot is immediately written to a bmp-image file
	named _bmp_target_name_ with 256 colors. no check for free disk
	space is done.

	the coordinates passed are (left upper corner) (right lower corner),
	all values in pixel, not bytes! the (_x1_, _y1_) must be at least
	one pixel right and below (_x0_, _y0_), and neither point may be
	off screen.

	_palette_ is always saved along with the image, since this is the
	only way it makes sense. (it'd also be pointless to try to convert
	it to a true color file.)

	return value is the number of bytes written, '0' in case of an
	error. */

int save_bmp_box(char *bmp_target_name, int x0, int y0, int x1, int y1,
	BYTE *palette){
	int	x, y;				/* loop variables */
	int	x_ext= x1-x0,	/* width and height of the screen region */
			y_ext= y1-y0;
	long	index;
	long	bmp_length= BMP_PALETTE_START + 1024 + (x_ext*y_ext);
								/* total file length */
	BYTE	*bmp_data,		/* buffer for bmp data */
			bmp_palette[1024];
								/* palette converted to bmp format */

	/* coordinate control: */
	if(x0<0 || x1>MAX_X || y0<0 || y1>MAX_Y || x1<=x0 || y1<=y0){
		/* something with the source bmp borders went wrong */
		x_set_text_mode();
		fprintf(stderr, "Illegal source borders when trying to save \
			screen to bmp format! \n");

		return (0);
	}
	else{
		/* customize bmp file header: */
		write_long_intel(&bmp_header[BMP_LENGTH], bmp_length);
			/* total space required */

		write_long_intel(&bmp_header[BMP_DATA_LENGTH], (bmp_length-1078));
			/* sheer image data length- according to Born, subtract
				file header offset and palette length from total file
				length. */

		write_long_intel(&bmp_header[BMP_WIDTH], x_ext);
		write_long_intel(&bmp_header[BMP_HEIGHT], y_ext);
			/* image width and height */

		/* allocate enough buffer space: */
		bmp_data= byte_array(bmp_length);

		/* convert palette to bmp format: */
		palette_make_bmp(bmp_palette, palette);

		/* copy header and palette to bmp buffer: */
		memcpy(bmp_data, bmp_header, BMP_PALETTE_START);
		memcpy(&bmp_data[BMP_PALETTE_START], bmp_palette, 1024);

		/* copy screen rectangle: */
		for(index=BMP_PALETTE_START+1024, y=y1-1; y>=y0; y--){
			/* row by row, bottom up! */
			for(x=x0; x<x1; x++){
				bmp_data[index++]= x_get_pixel(x, y, Page0_Offs);
			}
		}

		/* write to disk, as conveniently provided by Paul: */
		write_binary(bmp_target_name, 0, bmp_data, bmp_length);

		/* ... and clear memory: */
		free(bmp_data);

		return bmp_length;
	}
}
